from .computation_engine import apply, join, summarize, Engine, files_from, DelayedException
from .fs_cache import CacheBuilder, LocalCacheBuilder
from .std_logger import get_logger

# DO NOT TOUCH THE FOLLOWING LINE:
import sys
if not getattr(sys, "frozen", False):
    import pkg_resources  # part of setuptools
    __version__ = tuple(map(int, pkg_resources.require(__name__)[0].version.split(".")))
else:
    __version__ = (0, 0, 0)





def serialize_tuple(item):
    """serializes a tuple to a string.
       the tuple contains strings mixed with nested tuples which contain strings....

       the output uses brackets to denote tuple 'boundaries', "," is used to seperate
       the items of a tuple and a non-empty tuples conversion must end with a ","
       (so we can distinguis the conversion output of () and ("",) )

       a backslash escapes the interpetation of special charachters (brackets, ",", backslash)
    """
    if isinstance(item, tuple):
        if len(item) == 0:
            return "[]"
        return "[" + ",".join(serialize_tuple(t) for t in item) + ",]"
    assert isinstance(item, basestring), "item is of wrong type %r" % item
    item = item.replace("\\", r"\\")
    return item.replace("[", r"\[").replace("]", r"\]").replace(",", r"\,")


def _parse(ts, i=0):
    """helper function for reversing the output of `serialize_tuple`"""
    rv = []
    item = ""
    assert ts[0] == "["
    i += 1
    while i < len(ts):
        if ts[i] == "]":
            return tuple(rv), i + 1
        if ts[i] not in "[],":
            if ts[i] == "\\":
                i += 1
            item += ts[i]
            i += 1
            continue
        if ts[i] == "[":
            item, i = _parse(ts, i)
        if ts[i] == ",":
            rv.append(item)
            item = ""
            i += 1
            continue
        raise Exception("syntax error in '%s' at position %d" % (ts, i))


def unserialize_tuple(ts):
    """reverses the output of serialize_tuple.

    `unserialize_tuple` creates a (nested) tuple from a string which conforms the structure of the
    output of `serialize_tuple`.
    """
    return _parse(ts, 0)[0]



# encoding: utf-8

import os
import hashlib
import functools

import dill
import multiprocessing

from .std_logger import get_logger
from .serializer import serialize_tuple, unserialize_tuple


class CacheItem(object):

    def __init__(self, file_name, value, origin, loaded):
        self.file_name = file_name
        self.value = value
        self.origin = origin
        self.loaded = loaded

    def __str__(self):
        return str(self.value)


class _CacheBuilder(object):

    def __init__(self, root_dir=None, keep_in_memory=True):
        self.root_dir = root_dir
        self.keep_in_memory = keep_in_memory
        self.reset_handlers()


    def reset_handlers(self):
        self._type_handlers = list()

    def register_handler(self, type_, hash_function, file_extension, load_function, save_function,
                         get_origin_function, set_origin_function):
        self._type_handlers.append((type_,
                                    self._pickler(hash_function),
                                    file_extension,
                                    self._pickler(load_function),
                                    self._pickler(save_function),
                                    self._pickler(get_origin_function),
                                    self._pickler(set_origin_function),
                                    )
                                   )


    def __call__(self, function):
        folder = function.__name__
        if self.root_dir is not None:
            folder = os.path.join(self.root_dir, folder)

        cache, lock, counter, handlers = self._setup_cache_internals()

        c = _CachedFunction(function, folder, self.keep_in_memory, lock, cache, counter, handlers,
                            self._pickler, self._unpickler)
        return c


class CacheBuilder(_CacheBuilder):

    def __init__(self, root_dir=None, keep_in_memory=True):
        super(CacheBuilder, self).__init__(root_dir, keep_in_memory)
        self._pickler = dill.dumps
        self._unpickler = dill.loads
        self._manager = multiprocessing.Manager()

    def _setup_cache_internals(self):
        cache = self._manager.dict()
        lock = self._manager.Lock()
        counter = self._manager.Value('d', 0)
        handlers = self._manager.list(self._type_handlers)
        return cache, lock, counter, handlers


class LocalCounter(object):

    def __init__(self):
        self.value = 0


class NoOpContextManager(object):

    def __enter__(self, *a, **kw):
        pass

    __exit__ = __enter__


class LocalCacheBuilder(_CacheBuilder):

    """Cache which only resists in current process, can not be used with pacerd distributed
    computation capabilities ! Use CacheBuilder instead.
    """

    def __init__(self, root_dir=None, keep_in_memory=True):
        super(LocalCacheBuilder, self).__init__(root_dir, keep_in_memory)
        self._manager = None
        self._pickler = self._unpickler = lambda o: o


    def _setup_cache_internals(self):
        cache = dict()
        lock = NoOpContextManager()
        counter = LocalCounter()
        handlers = list(self._type_handlers)
        return cache, lock, counter, handlers


class _CachedFunction(object):

    """ Instances of this class can be used to decorate function calls for caching their
    results, even if the functions are executed across different processes started by Python
    multiprocessing modules Pool class.

    The cache is backed up on disk, so that cache entries are persisted over different
    runs.
    """

    def __init__(self, function, folder, _keep_in_memory, _lock, _cache, _counter, _handlers,
            _pickler, _unpickler):

        self.function = function
        self.__name__ = function.__name__
        self.folder = folder
        self._keep_in_memory = _keep_in_memory

        self._cache = _cache
        self._lock = _lock
        self._hit_counter = _counter
        self._type_handlers = _handlers

        self._pickler = _pickler
        self._unpickler = _unpickler

        self._logger = get_logger(self)
        self._setup_cache()

    def reset_handlers(self):
        del self._type_handlers[:]

    def register_handler(self, type_, hash_function, file_extension, load_function, save_function,
                         get_origin_function, set_origin_function):
        self._type_handlers.append((type_,
                                    self._pickler(hash_function),
                                    file_extension,
                                    self._pickler(load_function),
                                    self._pickler(save_function),
                                    self._pickler(get_origin_function),
                                    self._pickler(set_origin_function),
                                    )
                                   )

    def _lookup_for_type_of(self, obj):
        for row in self._type_handlers:
            if isinstance(obj, row[0]):
                return row
        nonf = self._pickler(None)
        return [None, nonf, None, nonf, nonf, nonf, nonf]

    def _lookup_hash_function_for(self, key):
        return self._unpickler(self._lookup_for_type_of(key)[1])

    def _lookup_load_function_for(self, ext):
        for (__, __, ext_i, load_function, __, __, __) in self._type_handlers:
            if ext == ext_i:
                return self._unpickler(load_function)
        return None

    def _lookup_ext_and_save_function_for(self, what):
        row = self._lookup_for_type_of(what)
        return row[2], self._unpickler(row[4])

    def _lookup_get_origin_function(self, what):
        return self._unpickler(self._lookup_for_type_of(what)[5])

    def _lookup_set_origin_function(self, what):
        return self._unpickler(self._lookup_for_type_of(what)[6])

    def get_number_of_hits(self):
        return self._hit_counter.value

    def _setup_cache(self):
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        for file_name in os.listdir(self.folder):
            stem, ext = os.path.splitext(file_name)
            hash_code, __, origin_str = stem.partition("--")
            origin = unserialize_tuple(origin_str)
            self._cache[hash_code] = CacheItem(file_name, None, origin, False)

    def clear(self):
        self._cache.clear()

    def _compute_hash(self, key):
        hash_fun = self._lookup_hash_function_for(key)
        if hash_fun is not None:
            return hash_fun(key)
        if isinstance(key, (tuple, list)):
            data = "".join(self._compute_hash(item) for item in key)
        elif isinstance(key, set):
            data = "".join(self._compute_hash(item) for item in sorted(key))
        elif isinstance(key, dict):
            data = "".join(self._compute_hash(item) for item in key.items())
        elif isinstance(key, basestring):
            data = key
        elif hasattr(key, "__dict__"):
            data = self._compute_hash(key.__dict__)
        elif isinstance(key, (bool, int, long, float,)):
            data = str(self._pickler(key))
        else:
            raise Exception("can not compute hash for %r" % (key,))
        if not isinstance(data, basestring):
            raise RuntimeError("implementation error: data should be str, but is %s" % type(data))
        muncher = hashlib.sha1()
        muncher.update(data)
        return muncher.hexdigest()

    def _contains(self, hash_code):
        return hash_code in self._cache.keys()

    def _get(self, hash_code):
        item = self._cache[hash_code]
        if not item.loaded:
            value = self._load(item.file_name)
            item.value = self._set_origin(value, item.origin)
            item.loaded = True
            self._cache[hash_code] = item
        return item.value

    def _put(self, hash_code_args, args, result, origin):
        try:
            origin_str = serialize_tuple(origin).replace("/", "%")\
                                                .replace("\\", "%")\
                                                .replace(":", "%")
        except Exception, e:
            raise Exception("can not pickle %r beause of: %s" % (result, e.message,))

        if len(origin_str) > 100:
            origin_str = "[TOO_LONG]"
        path = self._store(result, hash_code_args + "--" + origin_str)
        if self._keep_in_memory:
            self._cache[hash_code_args] = CacheItem(None, result, origin, True)
        else:
            self._cache[hash_code_args] = CacheItem(path, None, origin, False)

    def _load(self, file_name):
        path = os.path.join(self.folder, file_name)
        stem, f_ext = os.path.splitext(file_name)
        load_function = self._lookup_load_function_for(f_ext)
        if load_function is not None:
            obj = load_function(path)
        else:
            obj = dill.load(open(path, "rb"))
        self._logger.info("loaded %s" % path)
        return obj

    def _store(self, what, stem):
        ext, save_function = self._lookup_ext_and_save_function_for(what)
        if ext is not None and save_function is not None:
            path = os.path.join(self.folder, stem + ext)
            save_function(what, path)
        else:
            path = os.path.join(self.folder, stem + ".pickled")
            with open(path, "wb") as fp:
                dill.dump(what, fp)
        self._logger.info("stored %s" % path)
        return path

    def _get_origin(self, arg):
        get_origin = self._lookup_get_origin_function(arg)
        if get_origin is not None:
            origin = get_origin(arg)
            return origin if origin is not None else "XXX"
        if isinstance(arg, (tuple, list)):
            return tuple(self._get_origin(item) for item in arg)
        elif isinstance(arg, basestring):
            return arg
        elif isinstance(arg, (int, bool, float)):
            return str(arg)
        elif isinstance(arg, dict):
            return "dict(%d)" % len(arg)
        elif isinstance(arg, set):
            return "set(%d)" % len(arg)
        else:
            return "XXX"

    def _set_origin(self, obj, origin):
        set_origin = self._lookup_set_origin_function(obj)
        if set_origin is not None:
            return set_origin(obj, origin)
        return obj

    def __call__(self, *args, **kw):
        hash_code = self._compute_hash(args + tuple(sorted(kw.items())))
        if self._contains(hash_code):
            with self._lock:
                self._hit_counter.value += 1
            self._logger.info("cache hit for %s" % hash_code)
            return self._get(hash_code)
        result = self.function(*args, **kw)
        origin = self._get_origin(args)
        result = self._set_origin(result, origin)
        self._logger.info("new result for %s" % hash_code)
        self._put(hash_code, args, result, origin)
        return result



import multiprocessing
import dill

"""
This module is a non complete replacement for pythons multiprocessing module.
The existing multiprocessing module is restricted for remote execution of eg lambda functions,
decorated functions and static member functions.

We use dill for pickling instead of cPickle for transfering function code, arguments
and results.
"""


def dill_apply(fun_s, args_s, kwargs_s=None):
    """
    this function is executed in a separate process and decodes the input string
    into a functino and its args.
    """
    try:
        fun = dill.loads(fun_s)
        args = dill.loads(args_s)
        if kwargs_s is not None:
            kwargs = dill.loads(kwargs_s)
        else:
            kwargs = dict()
        res = fun(*args, **kwargs)
        return dill.dumps(res)
    except:
        import traceback
        traceback.print_exc()
        raise


class AsyncResult(object):

    def __init__(self, job):
        self.job = job

    def get(self):
        return dill.loads(self.job.get())


class SyncResult(object):

    def __init__(self, fun, args):
        self.fun = fun
        self.args = args
        self.res = self.fun(*self.args)

    def get(self):
        return self.res


class Pool(object):

    def __init__(self, n):
        self.n = n
        if self.n > 0:
            self.pool = multiprocessing.Pool(n)

    def apply_async(self, fun, args, kwargs=None):
        if self.n == 0:
            return SyncResult(fun, args)
        try:
            fun_c = dill.dumps(fun)
        except dill.PicklingError, e:
            raise Exception("%s\ndill failed to  pickle %s" % (e, fun))
        try:
            args_c = dill.dumps(args)
        except dill.PicklingError, e:
            raise Exception("%s\ndill failed to  pickle %s" % (e, args))
        kwargs_c = None
        if kwargs is not None:
            try:
                kwargs_c = dill.dumps(kwargs)
            except dill.PicklingError, e:
                raise Exception("%s\ndill failed to  pickle %s" % (e, kwargs))
        job = self.pool.apply_async(dill_apply, (fun_c, args_c, kwargs_c))
        return AsyncResult(job)



import tblib.pickling_support
# allows pickling of basic information from tracebacks:
tblib.pickling_support.install()

import abc
import collections
import dill
import fnmatch
import functools
import itertools
import multiprocessing
import os
import sys
import Queue
import random
import threading
import time
import traceback

from pacer.std_logger import get_logger


class DelayedException(object):

    """store exception for throwing it later.
       we use this for transfering an exception on a worker process to the
       main process.
       only works because we imported and installed tblib above
       """

    def __init__(self, ee):
        self.ee = ee
        __,  __, self.tb = sys.exc_info()

    def re_raise(self):
        raise self.ee, None, self.tb

    def as_string(self):
        return "".join(traceback.format_exception(self.ee, None, self.tb, limit=None))


def eval_pickled(f_s, args_s):
    """ requires function and args serialized with dill, so we can remote execute even lambdas or
    decorated functions or static methods, which 'pure' multiprocssing can not handle"""

    try:
        args = dill.loads(args_s)
    except Exception, e:
        get_logger(eval_pickled).error("got exception when unpickling args: {}".format(e))
        return Exception(e.message)
    try:
        f = dill.loads(f_s)
    except Exception, e:
        get_logger(eval_pickled).error("got exception when unpickling f: {}".format(e))
        return Exception(e.message)
    try:
        get_logger(f).info("start {}{} in process {}".format(f.__name__, args, os.getpid()))
        result = f(*args)
        get_logger(f).info("got result {!r} from process {}".format(result, os.getpid()))
        return result
    except Exception, e:
        get_logger(f).error("got exception: {}".format(e))
        return DelayedException(e)


class EnumeratedItem(object):

    def __init__(self, number, value):
        self.number = number
        self.value = value

    def __iter__(self):
        return iter((self.number, self.value))


def unpack(enumerated_args):
    new_enumeration = tuple(i for (i, __) in enumerated_args)
    if len(new_enumeration) == 1:
        new_enumeration = new_enumeration[0]
    argument = tuple(a for (__, a) in enumerated_args)
    return new_enumeration, argument


def wrap_callback(new_enumeration, callback):
    """creates a new callback which builds a EnumeratedItem from the reult """
    def new_callback(result):
        callback(EnumeratedItem(new_enumeration, result))
    return new_callback


class Engine(object):

    pool = None

    @staticmethod
    def set_number_of_processes(n):
        if n is None:
            Engine.pool = None
            return
        if n < 0:
            n = multiprocessing.cpu_count() - n
            n = max(n, 0)
        if n > 0:
            Engine.pool = multiprocessing.Pool(n)
        elif n == 0:
            Engine.pool = None
        else:
            raise Exception("can not set number of processes to %d" % n)

    @staticmethod
    def run_async(f, enumerated_args, callback):
        """ f is not implemented for handling enumerated items, so we exctract the items
        and construct a 'new enumeration' for the result.
        This is needed in order to ensure that the final result of the computations have
        a fixed order although their computation order might be different due to the
        concurrent processing."""
        assert all(isinstance(arg, EnumeratedItem) for arg in enumerated_args)
        for arg in enumerated_args:
            if isinstance(arg.value, DelayedException):
                callback(arg)
                return
        new_enumeration, args = unpack(enumerated_args)
        # dill can pickle more function types as pickle module (which underlies
        # multiprocessing) , eg lambdas or static methods, so we transport our computation job
        # serialized as strings, which we deserialize in the remote process again:
        f_s = dill.dumps(f)
        args_s = dill.dumps(args)
        # we use eval_pickled in both cases to make shure we have the same kind of error handling
        # for local and remote execution. this supports debugging.
        if Engine.pool is None:
            result = eval_pickled(f_s, args_s)
            wrap_callback(new_enumeration, callback)(result)
        else:
            Engine.pool.apply_async(eval_pickled, (f_s, args_s),
                                    callback=wrap_callback(new_enumeration, callback))


class DataStream(object):

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self._stream = Queue.Queue()

    def empty(self):
        return self._stream.empty()

    def get(self):
        return self._stream.get()

    def size(self):
        return self._size

    def put(self, item):
        """this is where computation results are passed to."""
        self._stream.put(item)

    def get_all_in_order(self, do_not_reraise_exceptions=False):
        results = [self.get() for _ in range(self.size())]
        if not do_not_reraise_exceptions:
            for result in results:
                if isinstance(result.value, DelayedException):
                    result.value.re_raise()
        results.sort(key=lambda en_item: en_item.number)
        return [item.value for item in results]


class Source(DataStream):

    def __init__(self, source):
        super(Source, self).__init__()
        self._size = len(source)
        for i, item in enumerate(source):
            self.put(EnumeratedItem(i, item))

    def __iter__(self):
        while not self.empty():
            yield self.get().value


def files_from(folder, *patterns):
    files = []
    for f in os.listdir(folder):
        if not patterns or any(fnmatch.fnmatch(f, pattern) for pattern in patterns):
            files.append(os.path.join(folder, f))
    return Source(files)


class Computation(DataStream):

    def __init__(self, f, inputs):
        super(Computation, self).__init__()
        self.f = f
        self.input_streams = self._setup_inputs(inputs)
        self._size = self.compute_size()

    def _setup_inputs(self, inputs):
        streams = []
        for input_ in inputs:
            stream = input_
            if not isinstance(input_, DataStream):
                if isinstance(input_, (tuple, list)):
                    stream = Source(input_)
                else:
                    stream = Source((input_,))
            streams.append(stream)
        return streams

    def start_computations(self):
        """
        starts computations at leafs first then up to the root
        """
        for stream in self.input_streams:
            if isinstance(stream, Computation):
                stream.start_computations()
        self._start_computation()
        return self

    @abc.abstractmethod
    def compute_size(self):
        pass

    @abc.abstractmethod
    def _start_computation(self):
        pass


class _UpdateManagerForJoin(object):

    """ manages input data from sources which might come in in arbitrary order.  handles
    incremental generation of input_streams for computations.  """

    def __init__(self, input_streams):
        self.seen = set()
        # a list of collected data per stream:
        self.data = [list() * a.size() for a in input_streams]

    def new_function_arguments_for(self, stream_index, item):
        self.data[stream_index].append(item)
        # cross product over all indices of available data:
        sizes = [len(v) for v in self.data]
        for perm in itertools.product(*(range(size) for size in sizes)):
            if perm not in self.seen:
                args = tuple(self.data[j][pi] for j, pi in enumerate(perm))
                yield args
                self.seen.add(perm)

    def input_data_pending(self, size):
        return len(self.seen) < size


class JoinComputation(Computation):

    def compute_size(self):
        return reduce(lambda x, y: x * y, (a.size() for a in self.input_streams))

    def _start_computation(self):

        manager = _UpdateManagerForJoin(self.input_streams)
        while manager.input_data_pending(self.size()):
            started_computation = False
            for i, stream in enumerate(self.input_streams):
                if not stream.empty():
                    item = stream.get()
                    for args in manager.new_function_arguments_for(i, item):
                        Engine.run_async(self.f, args, callback=self.put)
                        started_computation = True
            if not started_computation:
                time.sleep(0.001)


class SummarizeComputation(Computation):

    def __init__(self, f, inputs):
        assert len(inputs) == 1, "reduce computations only accept one input stream"
        super(SummarizeComputation, self).__init__(f, inputs)

    def compute_size(self):
        return 1

    def _start_computation(self):
        items = []
        stream = self.input_streams[0]
        for i in range(stream.size()):
            item = stream.get()
            if isinstance(item.value, DelayedException):
                self.put(item)
                return
            items.append(item)
        items.sort(key=lambda item: item.number)
        values = [v for __, v in items]
        arg = EnumeratedItem(0, values)
        Engine.run_async(self.f, (arg,), callback=self.put)


class ZipComputation(Computation):

    """either input_streams provide all n inputs or one fixed input. The latter is managed by
    ConstantSource class.
    """

    def compute_size(self):
        sizes = set(s.size() for s in self.input_streams)
        if 1 in sizes:
            sizes.remove(1)
        if len(sizes) > 1:
            msg = "allow ownly sources with same sized outputs or constant constant source"
            raise Exception(msg)
        if len(sizes) == 1:
            return sizes.pop()
        # sizes has size 0 means: we only had inputs of size 1
        return 1

    def _setup_computations(self):
        cip = self._collect_data_from_constant_input_streams()

    def _collect_data_from_constant_input_streams(self):
        constant_inputs = [None] * len(self.input_streams)
        for i, stream in enumerate(self.input_streams):
            if stream.size() == 1:
                constant_inputs[i] = stream.get()
        return constant_inputs

    def _check_for_any_new_data(self):
        for stream in self.input_streams:
            if stream.size() > 1:
                if not stream.empty():
                    return True
        return False

    def _assemble_function_arguments(self, collected_input_data, cip):
        for i, stream in enumerate(self.input_streams):
            if not stream.empty():
                item = stream.get()
                if item.number not in collected_input_data.keys():
                    collected_input_data[item.number] = cip
                collected_input_data[item.number][i] = item
        for key, arg in collected_input_data.items():
            if all(c is not None for c in arg):
                yield arg
            del collected_input_data[key]

    def _start_computation(self):
        n = 0
        # wait and get the constant args which we need anyway:
        cip = self._collect_data_from_constant_input_streams()
        collected_input_data = dict()
        # only constant inputs ?
        if self.size() == 1:
            Engine.run_async(self.f, cip, callback=self.put)
        else:
            while n < self.size():
                avail = self._check_for_any_new_data()
                if avail:
                    for args in self._assemble_function_arguments(collected_input_data, cip):
                        Engine.run_async(self.f, args, callback=self.put)
                        n += 1
                else:
                    time.sleep(0.001)


def apply(inner):
    @functools.wraps(inner)
    def wrapped(*args):
        return ZipComputation(inner, args)
    wrapped.inner = inner
    return wrapped


def join(inner):
    @functools.wraps(inner)
    def wrapped(*args):
        return JoinComputation(inner, args)
    wrapped.inner = inner
    return wrapped


def summarize(inner):
    @functools.wraps(inner)
    def wrapped(*args):
        return SummarizeComputation(inner, args)
    wrapped.inner = inner
    return wrapped


if __name__ == "__main__":

    @join
    def add(*args):
        time.sleep(0.5 * random.random())
        return sum(args)

    @apply
    def inc(x, config=dict()):
        time.sleep(0.5 * random.random())
        return x + config.get("increment", 1)

    @summarize
    def avg(values):
        return float(sum(values)) / len(values)

    Engine.set_number_of_processes(7)
    N = 3

    s1 = range(N)
    s2 = range(1, N + 1)
    r = add(0, 1, s1)

    r = inc(avg(inc(add(0, 1, add(inc(s1), inc(s2), inc(7))), dict(increment=2))))

    r.start_computations()

    print(sorted(r.get_all_in_order()))



import logging


def get_logger(obj=None):

    if obj is not None:
        if hasattr(obj, "__name__"):
            name = "%s.%s" % (obj.__module__, obj.__name__)
        else:
            module = obj.__class__.__module__
            clzname = obj.__class__.__name__
            name = "%s.%s" % (module, clzname)
    else:
        name = ""

    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    name = "%-45s" % name
    default_formatter = logging.Formatter("%(asctime)s:%(processName)-14s:" + name + ":%(levelname)s: %(message)s")

    if not(len(logger.handlers)):
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(default_formatter)
        logger.addHandler(console_handler)
    return logger
